;@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
;@                                                                            @
;@            S y m b O S   -   M S X   D e v i c e   D r i v e r             @
;@                OCM SD/MMC  INTERFACE  (for Symbos ver.2.1)                 @
;@                                                                            @
;@             (c) 2000-2007 by Prodatron / SymbiosiS (Jrn Mika)             @
;@                              2008 by caro                                  @
;@                                                                            @
;@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

;--- MAIN ROUTINES ------------------------------------------------------------
;### MSDOUT -> write sectors (512b)
;### MSDINP -> read sectors (512b)
;### MSDACT -> read and init media (only hardware- and partiondata, no filesystem)

;--- WORK ROUTINES ------------------------------------------------------------
;### MSDRED -> wait for data request and read 512byte
;### MSDWRT -> wait for data request and write 512byte
;### MSDERR -> check error state

;--- SUB ROUTINES -------------------------------------------------------------
;### MSDSEC -> converts logical sector number into physical sector number
;### MSDADR -> set sector address
;### MSDSHW -> maps MSD memory  to #4000-#7FFF

stobnkx equ #202	;memory mapping, when low level routines 
			;read/write sector data
bnkmonx equ #203	;set special memory mapping during mass storage access
bnkmofx equ #206	;reset special memory mapping during mass storage access
bnkdofx equ #209	;hide mass storage device rom
stoadrx equ #20c	;get device data record
clcd16x equ #20f	;HL=BC/DE, DE=BC mod DE
stobufx equ #212	;address of 512byte buffer
stoadry equ #214	;actual device

;==============================================================================
;### HEADER ###################################################################
;==============================================================================

org #1000-32
relocate_start

db "SMD2"               ;ID
dw msdend-msdjmp        ;code length
dw relocate_count       ;number of relocate table entries
ds 8                    ;*reserved*
db 1,0,0                ;Version Major, Version Minor, Type (0=FDC, 1=IDE, 2=SCSI)
db "OCM MegaSD   "  	;comment

msdjmp  dw msdinp,msdout,msdact,msdmof
msdmof  ret:dw 0
        db 1		;type/ID
        ds 4		;reserv
msdslt  dw 0		;slots parametrs
	db 0

;==============================================================================
;### MAIN ROUTINES ############################################################
;==============================================================================

;### MSDINP -> read sectors (512b)
;### Input      A=device (0-7), IY,IX=logical sector number, 
;### B=number of sectors, DE=address, (stobnkx)=banking config
;### Output     CF=0 -> ok
;###            CF=1 -> A=error code (...)
;### Destroyed  AF,BC,DE,HL,IX,IY
msdinp	or a
	jr z,dev0rd	;only A:
	xor a
	scf
	ret		;error
dev0rd	push de		;address
        call msdsec     ;c,d,e - sector in SD
        pop hl          ;HL=destination address
	push de
	ld a,(stobnkx)
	call bnkmonx
	ld (wp_rd+1),a
	call bnkmofx
	pop de
	jp rd_secs	;read B sectors
;
;### MSDOUT -> write sectors (512b)
;### Input      A=device (0-7), IY,IX=logical sector number, 
;		B=number of sectors, DE=address, (stobnkx)=banking config
;### Output     CF=0 -> ok
;###            CF=1 -> A=error code (...)
;### Destroyed  AF,BC,DE,HL,IX,IY
msdout	or a
	jr z,dev0wr	;only A:
	xor a
	scf
	ret		;error
dev0wr	push de		;DE=address
        call msdsec	;(c,d,e)*2 - address in SD
        pop hl          ;HL=source address
	push de
	ld a,(stobnkx)
	call bnkmonx
	ld (wp_wr+1),a
	call bnkmofx
	pop de
	jp wr_secs	;
;
;### MSDACT -> 	read and init media (only hardware- and partiondata, 
;		no filesystem)
;### Input      A=device (0-7)
;### Output     CF=0 -> ok
;###            CF=1 -> A=error code (...)
;### Destroyed  AF,BC,DE,HL,IX,IY
msdact	ld hl,0
	ld (n_sec+1),hl
rd_sec0	push af
	ld ix,0
        ld iy,0
        ld b,1		;1 sector
        ld de,(stobufx)	;address for read data
        call msdinp	;read first physical sector (=MBR)
	pop bc
        ret c           ;read failed -> error
; read OK
	ld hl,(stobufx)
	ld a,(hl)
	cp #EB
	jr z,yes_bpb
	or a
	jr z,yes_mbr
	xor a
	scf
	ret		;error
;
yes_mbr	ld de,#1C6
	add hl,de
	ld e,(hl):inc hl
	ld d,(hl)
	ld (n_sec+1),de
	ld a,b
	jr rd_sec0
;
yes_bpb	ld a,b
	call stoadrx	;get device data record (system)
        ld (hl),stotypoky ;=1 device ready
        inc hl
        ld (hl),stomedhdi ;=16 media type is MSD-HD
        ld de,stodatbeg-stodattyp
        add hl,de
	ld d,h:ld e,l:inc de
	ld (hl),0
	ld bc,3
	ldir
        xor a		;cy=0 OK
        ret
;==============================================================================
;### SUB ROUTINES #############################################################
;==============================================================================
;### MSDSEC -> converts logical sector number into physical sector number
;### Input      A=device (0-7), IY,IX=logical sector number
;### Output     C,D,E - number sector in SD
;### Destroyed  F,C,IX,IY
msdsec	ld a,b		;b=number
	push ix
	pop de
	push iy
	pop bc
	push hl
n_sec	ld hl,0
	add hl,de
	ex de,hl
	jr nc,no_pp
	inc c
no_pp	pop hl
	ld b,a
        ret
;
;### MSDSHW -> maps MSD memory  to #4000-#7FFF
;### Destroyed  A
msdshw 	di
	ld a,(msdslt+0)
        out (#a8),a
	ld a,(msdslt+1)
        ld (#ffff),a
	ld a,(msdslt+2)
        out (#a8),a
        ret
;===================================================
sd_off	push af
	xor a		;a=00h
	jr sd_set
;====
sd_on	push af
	ld a, #40
sd_set	ld (#6000), a
	pop af
	ret
;===================================================
;
set0CMD	ld de,0		;
	ld c,0		;
;
set_CMD	ld hl,#4000	;/CS=0 (a12=0)
	ld a, (hl)
	sla e
	rl d
	rl c
	ld (hl), b	; CMD
	nop
	ld (hl), c
	nop
	ld (hl), d
	nop
	ld (hl), e
	nop
	ld (hl), 0
	nop
	ld (hl), #95
	nop
	ld a, (hl)
	ld b, 16
L_7A8A	ld a, (hl)
	cp #FF
	ccf
	ret nc		;a<ff no error
	djnz L_7A8A
	scf		; error
	ret
;===================================================
init_SD	ld b, 10	;80 pulses CLK
L_7A98	ld a, (#5000)	; /CS=1 (a12=1)
	djnz L_7A98	;
;\----
	ld b, #40	; CMD=0 GO_IDLE_STATE
	call set0CMD
	ret c		; error
	and #F7		;
	cp 1		;bit0 - in idle state ?
	scf
	ret nz		;error
; bit 1=1 (in idle state)
L_7AAF	ld b, #77	; CMD=55 APP_CMD
	call set0CMD
	and 4		;bit 2 - illegal command
	jr z, L_7ACD	;error command
; command OK
; for MMC-card
	ld b, #41	; CMD=1 SEND_OP_COND
	call set0CMD
	jr L_7AD9	;
;--
; for SD-card
L_7ACD	ld b, #69	; CMD=41 APP_SEND_OP_COND
	call set0CMD
L_7AD9	ret c		; error
	bit 0,a		; idle state=1 ?
	jr nz, L_7AAF	; repeat
	or a
	ret z		; no error
	scf
	ret		; error
;---------------------------------------------------
rd_ini	call init_SD
	pop bc
	pop de
	pop hl
	jr nz,error0
	jr c, error0	; error
;===================================================
; INPUT:
; c,d,e - sector
; b - number
; hl - adress
; OUTPUT:
; CY=1 - error
rd_secs	push hl		;address
	push de		;c,d,e - sector
	push bc		;b=number
	call msdshw	;set page MegaSD
	call sd_on	;MegaSD on
	ld b, #51	; CMD=17 - read	single block
	call set_CMD
	jr c,rd_ini	; error
;
	pop bc		;b=number
	pop de		;c.d.e - sector
	pop hl		;hl=address
	or a
	jr nz, error0	; error
;
wp_rd	ld a,0		; change
	out (#FE),a
	push de
	push bc
	ex de, hl	;de=address
	ld bc, #200
	ld hl, #4000
rd_wait	ld a, (hl)
	cp #FE		;Data token
	jr nz,rd_wait
;\------
	ldir		; read sector 512 byte
	ld a, (hl)	;CRC1
	ld a, (hl)	;CRC2
	ex de, hl	;hl=address+200h
	call sd_off	;MegaSD off
	call bnkmofx	;set page2=1
	call bnkdofx	;set (#A8)=#FF;(#FFFF)=#AA
	pop bc
	pop de
	inc de		;next sector
	ld a, d
	or e
	jr nz,no_inc1
	inc c
no_inc1	djnz rd_secs
;\-----
	xor a		;no errors
	ret
;---------------------------------------------------
err_6	pop bc
	pop de
error0	call sd_off	;MegaSD off
	call bnkdofx	;set (#A8)(#FFFF)
	xor a
	scf		;error 0
	ret
;---------------------------------------------------
wr_ini	call init_SD
	pop bc
	pop de
	pop hl
	jr nz, error0
	jr c,  error0
;===================================================
; INPUT:
; c,d,e - sector
; b - number
; hl - adress
; OUTPUT:
; CY=1 - error
wr_secs	push hl		;adress
	push de		;c,d,e - sector
	push bc		;b=number
	call msdshw	;Set page MegaSD
	call sd_on	;MegaSD on
	ld b, #58	; CMD=24 write block
	call set_CMD
	jr c,wr_ini	; error
;       
	pop bc
	pop de
	pop hl		;adress
	or a
	jr nz, error0	; error
;
wp_wr	ld a,0
	out (#FE),a
	push de
	push bc
	ld bc, #200
	ld de, #4000
	ld a, (de)
	ld a, #FE
	ld (de), a
	ldir		; write	sectors	512 byte
; hl = address+200h
	ex (sp), hl
	ld a, (de)
	ex (sp), hl
	ld a, (de)
	ld a, (de)
	ld a, (de)
	and #1F
	cp 5
	jr nz,err_6	; error 6
;
wr_wait	ld a, (de)
	cp #FF
	jr nz,wr_wait
;\-----
	call sd_off	;MegaSD off
	call bnkmofx	;set page2=1
	call bnkdofx	;set (#A8)(#FFFF)
	pop bc
	pop de
	inc de		;next sector
	ld a, d
	or e
	jr nz,no_inc2
	inc c
no_inc2	djnz wr_secs	;next cikl
;\------
	xor a		; no error
	ret
;---------------------------------------------------
msdend

relocate_table
relocate_end
